/*
 * Functions for reading and converting various restricted
 * C64/etc and/or indexed/paletted graphics formats.
 * Programmed and designed by Matti 'ccr' Hamalainen
 * (C) Copyright 2012-2015 Tecnic Software productions (TNSP)
 *
 * Please read file 'COPYING' for information on license and distribution.
 */
#ifndef LIB64GFX_H
#define LIB64GFX_H 1

#include "libgfx.h"

#ifdef __cplusplus
extern "C" {
#endif


// Bitmap constants
#define C64_SCR_WIDTH          320
#define C64_SCR_HEIGHT         200
#define C64_SCR_CH_WIDTH       (C64_SCR_WIDTH/8)
#define C64_SCR_CH_HEIGHT      (C64_SCR_HEIGHT/8)
#define C64_SCR_COLOR_SIZE     (C64_SCR_CH_WIDTH * C64_SCR_CH_HEIGHT)
#define C64_SCR_SCREEN_SIZE    (C64_SCR_CH_WIDTH * C64_SCR_CH_HEIGHT)
#define C64_SCR_BITMAP_SIZE    ((C64_SCR_WIDTH * C64_SCR_HEIGHT)/8)
#define C64_SCR_EXTRADATA      1024
#define C64_SCR_MAX_BANK       8

// C64 video screen pixel aspect ratio on PAL
#define C64_SCR_PAR_XY         (0.9365f)

// Sprite constants
#define C64_SPR_WIDTH          3 // bytes
#define C64_SPR_HEIGHT         21 // lines
#define C64_SPR_WIDTH_PX       (8 * C64_SPR_WIDTH)
#define C64_SPR_SIZE           ((C64_SPR_WIDTH * C64_SPR_HEIGHT) + 1)

// Character constants
#define C64_CHR_WIDTH          1 // bytes
#define C64_CHR_HEIGHT         8 // lines
#define C64_CHR_WIDTH_PX       (8 * C64_CHR_WIDTH)
#define C64_CHR_SIZE           (C64_CHR_WIDTH * C64_CHR_HEIGHT)

// Etc.
#define C64_RAM_SIZE           (64*1024)
#define C64_NCOLORS            16
#define C64_MAX_COLORS         16
#define C64_VIDBANK_SIZE       (16*1024)
#define C64_MAX_SPRITES        (C64_VIDBANK_SIZE / C64_SPR_SIZE)
#define C64_MAX_CHARS          256


// Different supported C64 bitmap "modes"
enum
{
    D64_FMT_HIRES        = 0x0000,
    D64_FMT_MC           = 0x0001,
    D64_FMT_ILACE        = 0x0002,
    D64_FMT_FLI          = 0x0004,
    D64_FMT_CHAR         = 0x0008,

    D64_FMT_MODE_MASK    = 0x000f,

};

// Number of video banks
enum
{
    D64_FLI_2BANK,
    D64_FLI_4BANK,
    D64_FLI_8BANK,
};

enum
{
    D64_ILACE_COLOR,
    D64_ILACE_RES,
};


// Charmode screen memory configuration
enum
{
    D64_CHCFG_SCREEN,      // Use screen memory
    D64_CHCFG_LINEAR = 1,  // Generate linear pattern so that first line
                           // of chars is 0-39, second 40-79, up to 255.
};


typedef struct
{
    BOOL multicolor, xexpand, yexpand;
    int color, xc, yc;
    Uint8 data[C64_SPR_HEIGHT][C64_SPR_WIDTH];
} DMC64Sprite;


enum
{
    DC_D020,
    DC_BGCOL,
    DC_D021,
    DC_D022,
    DC_D023,
    DC_D024,
};


typedef struct
{
    int type,       // Image type (D64_FMT_*)
        fliType,    // FLI type (if FLI used)
        laceType,   // Interlace type (D64_ILACE_*)
        laceBank1,  // Interlace video bank indices
        laceBank2;

    int width, height; // Width and height in pixels
    int ch_width, ch_height; // Width and height in charblocks

    Uint8
        *color[C64_SCR_MAX_BANK],
        *bitmap[C64_SCR_MAX_BANK],
        *screen[C64_SCR_MAX_BANK],
        *charmem[C64_SCR_MAX_BANK],
        extradata[C64_SCR_EXTRADATA],
        d020, bgcolor, d022, d023, d024;

    DMC64Sprite sprites[C64_MAX_SPRITES];
} DMC64Image;


enum
{
    DT_COLOR_RAM,
    DT_BITMAP,
    DT_SCREEN_RAM,
    DT_COLOR_REG,
    DT_COLOR_SET,
    DT_EXTRADATA,

    DT_DEC_FUNCTION,
    DT_ENC_FUNCTION,

    DT_CHAR_DATA,
    DT_CHAR_CONFIG,

    DT_LAST,
};

#define D64_MAX_ENCDEC_OPS 16


typedef struct _DMC64EncDecOp
{
    int    type;
    size_t offs;
    int    bank;
    size_t size;
    BOOL   (*decfunction)(DMC64Image *img, const struct _DMC64EncDecOp *op, const Uint8 *buf, const size_t len);
    BOOL   (*encfunction)(const struct _DMC64EncDecOp *op, Uint8 **buf, size_t *len, const DMC64Image *img);
} DMC64EncDecOp;


typedef struct _DMC64ImageFormat
{
    int  type;   // Type flags, see D64_FMT_*
    char *fext;  // Filename extension
    char *name;  // Format description/name

    size_t addr; // Loading address (0 if no loading address)
    size_t size; // Size, including loading address. Only used in encoding, if even there (0 if no static size)

    int width, height; // Width and height in pixels
    int ch_width, ch_height; // Width and height in charblocks

    int  (*probe)(const Uint8 *buf, const size_t len, const struct _DMC64ImageFormat *fmt);
    int  (*decode)(DMC64Image *img, const Uint8 *buf, const size_t len, const struct _DMC64ImageFormat *fmt);
    int  (*encode)(DMC64Image *img, Uint8 **buf, size_t *len, const struct _DMC64ImageFormat *fmt);
    int  (*convertFrom)(DMImage *, const DMC64Image *);
    int  (*convertTo)(DMImage *, DMC64Image *);

    int nencdecOps;
    DMC64EncDecOp encdecOps[D64_MAX_ENCDEC_OPS];
} DMC64ImageFormat;


extern DMColor           dmC64Palette[C64_NCOLORS];
extern const DMC64ImageFormat  dmC64ImageFormats[];
extern const int         ndmC64ImageFormats;


DMC64Image *dmC64ImageAlloc(int width, int height, int ch_width, int ch_height);
void      dmC64ImageFree(DMC64Image *img);


char *    dmC64GetImageTypeString(char *buf, const size_t len, const int type);
int       dmC64ConvertCSDataToImage(DMImage *img, int xoffs, int yoffs, const Uint8 *inBuf, int width, int height, BOOL multicolor, int *colors);

int       dmC64DecodeGenericBMP(DMC64Image *img, const Uint8 *buf, const size_t len, const DMC64ImageFormat *fmt);
int       dmC64EncodeGenericBMP(Uint8 **pbuf, size_t *plen, const DMC64Image *img, const DMC64ImageFormat *fmt);
int       dmC64ConvertGenericBMP2Image(DMImage *dst, const DMC64Image *src);

int       dmC64ConvertBMP2Image(DMImage **pdst, const DMC64Image *src, const DMC64ImageFormat *fmt);
int       dmC64ProbeBMP(const Uint8 *buf, const size_t len, const DMC64ImageFormat **fmt);
int       dmC64DecodeBMP(DMC64Image **img, const Uint8 *buf, const size_t len, const size_t probeOffs, const size_t loadOffs, const DMC64ImageFormat **fmt, const DMC64ImageFormat *forced);


#ifdef __cplusplus
}
#endif

#endif // LIB64GFX_H
